# -*- coding: utf-8 -*-

""" This template outlines the necessary components for implementing a new Cell
in SEEDS.  Generally, this involves implementing a constructor (__init__) and
an update method.

Areas marked with 'TODO' should be replaced with code specific to the Cell
being implemented.

The name of the Cell type should be the name of the class.  The name of the
file should also match this.

Methods and parameters common to all Cell objects can be seen in the Cell.py
file in the main SEEDS codebase.

Once completed, new Cell type files can be placed in a plugins directory and
used by specifying the name of the file/class with the "cell" parameter in the
[Population] section of the configuration file.

"""

__author__ = "TODO"
__credits__ = "TODO"

import random

from seeds.Cell import *
from seeds.SEEDSError import *


class TODO-CellTypeName(Cell):

    """ TODO - some documentation about this Cell, the motivation behind it,
    and the configuration parameters it accepts.

    """

    # TODO: the "types" list specifies the different types of Cell that this
    # model supports.

    # TODO: the "max_types" property describes the maximum number of types
    # possible with this cell type.  Generally, this is equal to the length of
    # the "types" list, however it is possible to have more complex Cell types.
    # For example, one could assign a type for all unique genomes in a genetic
    # algorithm (GA) model.

    # TODO: the "type" colors list specifies the colors that should be used to
    # represent each Cell type in actions that produce images or movies.  The
    # length of this list should be equal to the length of the "types" list.

    types = ['Rock', 'Paper', 'Scissors']
    max_types = 3
    type_colors = ['r','g','b']

    ROCK = 0
    PAPER = 1
    SCISSORS = 2

    # TODO: the __init__ method (or "constructor") is called when a new Cell
    # object is created.  Generally, a Cell type will want to be assigned
    # during this step as well as any other properties associated with this
    # Cell.  This is often the stage where configuration values are read.

    def __init__(self, experiment, population, id, type=-1,
                 name="TODO-CellTypeName", label=None):
        """
        TODO: documentation
        """

        # Call the Cell constructor, assigning properties common to all cells.
        super(RPSCell, self).__init__(experiment, population, id, name=name, label=label)

        # Assign the organism the specified type, or pick one at random
        if type == -1:
            self.type = random.randint(0, len(self.types)-1)
        else:
            self.type = type
        
        # Keep track of how many organisms there are of this type.  You will
        # probably want to keep this code.
        self.population.increment_type_count(self.type)

        # TODO: read any configuration values


    # TODO: the __str__ method returns a string to be used when an object is
    # printed.  This can be a useful place to return information that is
    # specific to the Cell type
    def __str__(self):
        """Produce a string to be used when the object is printed"""
        return 'TODO-CellTypeName %d Type %d (%s)' % (self.id, self.type, self.types[self.type])

    def type(self):
        """Return the name of the type of this cell"""
        return self.types[self.type]


    # TODO: the update method updates an organism's state.  As such, it is the
    # most important part of a Cell object.  When a Cell is updated, it may
    # update its state based on environmental conditions, the composition of
    # its neighborhood, or through stochastic processes.  This dependes
    # completely on the type of system being modeled.

    def update(self):
        """
        TODO: documentation about how a Cell is updated
        """

        # If state is updated based on the composition of the neighborhood,
        # this code gets a list of neighboring Cell objects.
        neighbors = self.get_neighbors()

        # If a Cell's state depends on the level of some resource, at that
        # point in space, the following sample code gets that resource and sets
        # its value.
        res = self.experiment.get_resource("RESOURCE_NAME")
        res_cell = res.topology.get_nearest_node(coords=self.coords(), n=1)
        print("Current Resource Level: %f" % (res_cell.level))

        res_cell.level = max(0, res_cell.level - 1) # Consume up to 1 unit of resource


        # If the type of the Cell changes, the update_type_count method should
        # be called, which specifies the old type and the new type.
        self.population.update_type_count(OLD_TYPE, NEW_TYPE)




# -*- coding: utf-8 -*-
"""
A Cell object represents a place in the environment where an organism could
reside.  If that Cell is occupied by an organism, the Cell object also defines
that organism.
"""

__author__ = "Brian Connelly <bdc@msu.edu>"
__credits__ = "Brian Connelly"


class Cell(object):
    """
    Interface for Cell objects

    Properties:

    experiment
        A reference to the Experiment in which the Cell exists
    population
        A reference to the Population in which the Cell exists
    id
        A unique ID representing that Cell.  This should be the same as the
        node id in the population topology graph.
    types
        List of strings describing the possible types the Cell could be
    type
        Number indicating which type the current Cell is.  This number is also
        an index into the 'types' parameter.
    max_types
        Maximum number of different types possible with this Cell
    type_colors
        A list of colors (matplotlib color strings or hex colors) to be used to
        represent the different cell types by scripts that plot or visualize
        the population.  A default list is defined that allows for coloring of
        up to 8 types.
    name
        The name of the Cell type
    label
        A unique label identifying this Cell's configuration

    Configuration:
        Configuration options for each custom Cell object should be stored in a
        configuration block bearing the name of that Cell type (e.g.,
        "[DemoCell]")

    """

    def __init__(self, experiment, population, id, name=None, label=None):
        """Initialize a Cell object

        Parameters:
        
        *experiment*
            A reference to the Experiment in which this Cell exists
        *population*
            A reference to the Population in which this Cell exists
        *id*
            A unique ID for this cell
        *name*
            The name of the Cell type
        *label*
            A unique label for the configuration of this cell

        """

        self.experiment = experiment
        self.population = population
        self.id = id
        self.name = name
        self.label = label
        self.type = None
        self.max_types = 0
        self.type_colors = ['r','g','b','y','c', 'm', 'k']

        if self.label:
            self.config_section = "%s:%s" % (self.name, self.label)
        else:
            self.config_section = "%s" % (self.name)

    def __str__(self):
        """Produce a string to be used when a Cell object is printed"""
        return 'Cell %d Type %d' % (self.id, self.type)

    def add_neighbor(self, neighbor):
        """Make the given cell a neighbor"""
        self.population.topology.add_edge(self.id, neighbor.id)

    def remove_neighbor(self, neighbor):
        """Disconnect the Cell from the given Cell, making them no longer
        neighbors
        """
        self.population.topology.remove_edge(self.id, neighbor.id)

    def get_neighbors(self):
        """Get a list of neighboring cells"""
        return [self.population.topology.graph.node[n]['cell'] for n in self.population.topology.get_neighbors(self.id)]

    def update(self):
        """Update the Cell according to its update rules"""
        pass

    def teardown(self):
        """Perform any necessary cleanup at the end of the experiment"""
        pass

    def coords(self):
        """Get the coordinates of the Cell in space"""
        return self.population.topology.graph.node[self.id]['coords']

    def get_neighbor_distance(self, neighbor):
        """Get the Cartesian distance to the given neighbor Cell"""
        return self.population.cell_distance(self, neighbor)

    def get_neighbor_distances(self):
        """Get an array of distances to all neighbors"""
        return [self.get_neighbor_distance(n) for n in self.get_neighbors()]



# -*- coding: utf-8 -*-
"""This file defines several Exceptions which may be thrown by SEEDS."""

__author__ = "Brian Connelly <bdc@msu.edu>"
__credits__ = "Brian Connelly"


class SEEDSError(Exception):
    """Base class for all Exceptions related to SEEDS"""
    pass

class ResourceNotDefinedError(SEEDSError):
    """Error to be raised when a Resource is requested that has not been
    defined.

    Attributes:

    *resource*
        The name of the requested Resource (string)

    """

    def __init__(self, resource):
        self.resource = resource

    def __str__(self):
        return "Resource '%s' not defined" % (self.resource)


class PluginNotFoundError(SEEDSError):
    """Error to be raised when a Plugin (of any type) is not found

    Attributes:

    *plugin*
        The name of the plugin requested (string)
    """

    def __init__(self, plugin):
        self.plugin = plugin

    def __str__(self):
        return "Plugin '%s' not found" % (self.plugin)


class ActionPluginNotFoundError(PluginNotFoundError):
    """Error to be raised when a Action Plugin is not found

    Attributes:

    *action*
        The name of the Action requested (string)
    """

    def __init__(self, action):
        self.action = action

    def __str__(self):
        return "Action '%s' not found" % (self.action)


class CellPluginNotFoundError(PluginNotFoundError):
    """Error to be raised when a Cell Plugin is not found

    Attributes:

    *cell*
        The name of the cell requested (string)
    """

    def __init__(self, cell):
        self.cell = cell

    def __str__(self):
        return "Cell type '%s' not found" % (self.cell)


class ResourceCellPluginNotFoundError(PluginNotFoundError):
    """Error to be raised when a ResourceCell Plugin is not found

    Attributes:

    *resource*
        The name of the ResourceCell requested (string)
    """

    def __init__(self, resource):
        self.resource = resource

    def __str__(self):
        return "ResourceCell '%s' not found" % (self.resource)


class TopologyPluginNotFoundError(PluginNotFoundError):
    """Error to be raised when a Topology Plugin is not found

    Attributes:

    *topology*
        The name of the Topology requested (string)
    """

    def __init__(self, topology):
        self.topology = topology

    def __str__(self):
        return "Topology type '%s' not found" % (self.topology)


class InvalidParameterValue(SEEDSError):
    """Error to be raised when a given parameter value is invalid

    Attributes:

    *section*
        The name of the section in which the parameter is defined
    *parameter*
        The name of the parameter
    """

    def __init__(self, section, parameter):
        self.section = section
        self.parameter = parameter

    def __str__(self):
        return "Invalid value for parameter '%s.%s'" % (self.section, self.parameter)

class NonExistentNodeError(SEEDSError):
    """Error to be raised when a node does not exist

    Attributes:

    *id*
        The ID of the node
    """

    def __init__(self, id):
        self.id = id

    def __str__(self):
        return "Node %d does not exist in Topology" % (self.id)

class NonExistentEdgeError(SEEDSError):
    """Error to be raised when an edge does not exist

    Attributes:

    *src*
        The ID of the first node connected with the edge
    *dest*
        The ID of the second node connected with the edge
    """

    def __init__(self, src, dest):
        self.src = src
        self.dest = dest

    def __str__(self):
        return "Edge %d-%d does not exist in Topology" % (self.src, self.dest)

class ConfigurationError(SEEDSError):
    """Error to be raised when an invalid configuration is given, either
    through one bad parameter value or through parameter value conflicts.

    Attributes:

    *message*
        The message to be displayed (string)

    """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message




# -*- coding: utf-8 -*-
""" Cell type modeling the Conway's classic Game of Life cellular automaton.
Each cell is either "Alive" or "Dead", and when a cell is updated, it follows
the following rule:

- Live cells with less than 2 neighbors die (under-population)
- Live cells with 2 or 3 neighbors live
- Live cells with more than 3 neighbors die (overcrowding)
- Dead cells with 3 neighbors becomes alive (birth)

"""

__author__ = "Brian Connelly <bdc@msu.edu>"
__credits__ = "Brian Connelly"

import random

from seeds.Cell import *
from seeds.SEEDSError import *


class GameOfLifeCell(Cell):
    """
    This is a simple Cell type modeling Conway's Game of Life.

    This Cell type has 2 states:
        ALIVE: Cells that are alive
        DEAD: Cells that are dead

    """

    types = ['Alive', 'Dead']
    type_colors = ['k','w']
    max_types = 2

    ALIVE = 0
    DEAD = 1

    def __init__(self, experiment, population, id, type=-1, name="GameOfLifeCell", label=None):
        """Initialize a GameOfLifeCell object

        The type for the cell is selected at random.

        Parameters:

        *experiment*
            A reference to the Experiment
        *population*
            A reference to the Population in which this Cell exists
        *id*
            A unique ID for the cell
        *type*
            The type of cell to initialize (-1 for random)
        *name*
            The name of this Cell type
        *label*
            A unique label for configuring this Cell type

        """

        super(GameOfLifeCell, self).__init__(experiment, population, id, name=name, label=label)

        if type == -1:
            self.type = random.randint(0, len(self.types)-1)
        else:
            self.type = type
        
        self.population.increment_type_count(self.type)

    def __str__(self):
        """Produce a string to be used when the object is printed"""
        return 'GameOfLifeCell %d Type %d (%s)' % (self.id, self.type, self.types[self.type])

    def type(self):
        """Return the name of the type of this cell"""
        return self.types[self.type]

    def update(self):
        """Update the cell based on the following rules:

        - Live cells with less than 2 neighbors die (under-population)
        - Live cells with 2 or 3 neighbors live
        - Live cells with more than 3 neighbors die (overcrowding)
        - Dead cells with 3 neighbors becomes alive (birth)

        """

        neighbors = self.get_neighbors()

        if len(neighbors) < 1:
            print("Warning: Can not update GameOfLifeCell with 0 neighbors")
            return

        num_live_neighbors = 0
        num_dead_neighbors = 0

        for n in neighbors:
            if n.type == self.ALIVE: num_live_neighbors += 1
            elif n.type == self.DEAD: num_dead_neighbors += 1

        if self.type == self.ALIVE and num_live_neighbors < 2:
            self.type = self.DEAD
            self.population.update_type_count(self.ALIVE, self.DEAD)            
        elif self.type == self.ALIVE and num_live_neighbors > 3:
            self.type = self.DEAD
            self.population.update_type_count(self.ALIVE, self.DEAD)            
        elif self.type == self.DEAD and num_live_neighbors == 3:
            self.type = self.ALIVE
            self.population.update_type_count(self.DEAD, self.ALIVE)            



# -*- coding: utf-8 -*-
""" 
This Cell type is an implementation of the classic quasispecies model with
two peaks, a high but narrow peak, and a lower but wider peak. We implement
this cell type as a bit string, where the first bit represents if the organism
is on the wide peak or not (0 = narrow, 1 = wide) and the remaining bits dictate
where in the fitness landscape the genotype is. The more bits set to 1, the higher
the fitness. 

These two peaks are implemented as a polynomial function, where the order of the
function determines the steepness of the peak (i.e. x^3 is less steep than x^5).
The wide peak is a linear function where fitness varies from 0, to wide_max_value
with the number of bits set to 1. 

The classic quasispecies result is that when mutation rates are low, the quasispecies 
occupies the narrow peak since it has higher fitness. However, with higher mutation
rates the average fitness of a quasispecies is lower due to genotypes "falling" off
the peak. In these high mutation rate environments, the quasispecies is expected to
occupy the lower, but wider peak.

In this implementation, all cells die at the same rate. However, cells that replace
the empty space are chosen proportional to fitness (roulette wheel selection). When 
organisms reproduce, there is a per-site probability of flipping a bit, thus mutating
the genotype. 

"""

__author__ = "Luis Zaman <zamanlui@msu.edu>"
__credits__ = "Luis Zaman, Brian Connelly"

from seeds.Cell import *
import random

class QuasispeciesCell(Cell):
    """
    This cell type is an implementation of a quasispecies model using
    bitstrings.  See the top-level comment block for a more detailed
    explanation, also see Martin Nowak's Evolutionary Dynamics (chapter 3) for
    a slightly more in depth review. 
    
    The following config options should be specified in a [QuasispeciesCell]
    section of the seeds.cfg file:
    
        death_rate: death rate of cells (float, [0,1])
        genotype_length: number of bits the genotypes use (int, [2,])
        site_mut_rate: site_mut_rate: probability of a site mutating (applied per site) (float, [0,])
        narrow_polynomail_order: the order of the polynomial defining the narrow peak (int, [1,])
        wide_max_value: 
        
    Example - Narrow Favored:
        [QuasispeciesCell]
        death_rate = 0.2
        genotype_length = 20
        site_mut_rate = 0.005
        narrow_polynomail_order = 4
        wide_max_value = 0.75
        
    Example - Wide Favored:
        [QuasispeciesCell]
        death_rate = 0.2
        genotype_length = 20
        site_mut_rate = 0.1
        narrow_polynomail_order = 4
        wide_max_value = 0.75

    """

    types = ['Empty', 'Narrow', 'Wide']
    type_colors = ['#777777','b','r']
    max_types = 3

    EMPTY = 0
    NARROW = 1
    WIDE = 2

    def __init__(self, experiment, population, id, type=-1, name="QuasispeciesCell", label=None):
        """Initialize a QuasispeciesCell object

        The type for the cell is selected at random.

        Parameters:

        *experiment*
            A reference to the Experiment
        *population*
            A reference to the Population in which the Cell exists
        *id*
            A unique ID for the cell
        *type*
            The type of cell to initialize (-1 for random)
        *name*
            The name of this Cell type
        *label*
            A unique label for configuring this Cell type

        """

        super(QuasispeciesCell, self).__init__(experiment, population, id, name=name, label=label)

        self.death_rate = self.experiment.config.getfloat(self.config_section, 'death_rate')
        self.genotype_length = self.experiment.config.getint(self.config_section, 'genotype_length')
        self.site_mut_rate = self.experiment.config.getfloat(self.config_section, 'site_mut_rate')
        self.narrow_polynomail_order = self.experiment.config.getfloat(self.config_section, 'narrow_polynomail_order')
        self.wide_max_value = self.experiment.config.getfloat(self.config_section, 'wide_max_value')
        
        #make sure all of the parameters are okay
        assert self.death_rate >= 0
        assert self.genotype_length > 1
        assert self.site_mut_rate >= 0
        assert self.narrow_polynomail_order > 0
        assert self.wide_max_value >= 0

        #generate a random genotype
        self.genotype = [random.randint(0,1) for i in range(self.genotype_length)]
        
        if type == -1:
            #determine type from bitstring genotype, we'll say that 0 = narrow and 1 = wide
            #that way we can just add one to get our defined types 
            self.type = random.randint(0,len(self.types)-1)
        else:
            self.type = type
            
            #don't let this bit go negative... even though it will only happen wwith empty types
            #and they don't technically have genotypes anyway
        
        #set first bit of genotype appropriately 
        self.genotype[0] = max(self.type-1,0)
        
        self.world.increment_type_count(self.type)
        
    def flip_bit(self, bit):
        """Helper function to handle single bit mutations"""
        if bit == 0:
            return 1
        else:
            return 0
            
    def mutate(self, genotype):
        """Mutate genotype based on mutation rate"""
        new_genotype = []
        for bit in genotype:
            if random.random() < self.site_mut_rate:
                new_genotype.append(self.flip_bit(bit))
            else:
                new_genotype.append(bit)
        
        return new_genotype
        
      
    def get_fitness(self, genotype):
        """ Calculate fitness based on the number of bits set to 1 and the peak
        the organism is on.

        """
        
        #get % of genotype that is set to 1
        genotype_perc_one = genotype[1::].count(1)/float(len(genotype[1::]))
        
        if genotype[0] == 0:
            #raise it to the specified order to get fitness value for narrow type
            return genotype_perc_one**self.narrow_polynomail_order
        else:
            #use linear for wide type
            return genotype_perc_one*self.wide_max_value
            
            
    def choose_neighbor(self, orgs):
        """Do roulettle wheel selection between passed organisms (neighbors)
        and return winner

        """

        fitnesses = [self.get_fitness(o.genotype) for o in orgs]
        #add small amount to avoid division by 0
        sum_fitness = sum(fitnesses) + 0.0000001
        #would be faster with numpy arrays!
        norm_fitnesses = [f/float(sum_fitness) for f in fitnesses]
        
        #roll the ball, see where it falls
        r = random.random()
        partial_sum = 0
        for i in range(len(norm_fitnesses)):
            if partial_sum <= r <= partial_sum + norm_fitnesses[i]:
                return orgs[i]
            partial_sum += norm_fitnesses[i] 
        
        print("didn't find a neighbor... ", sum_fitness)
        return orgs[-1]
        
    def __str__(self):
        """Produce a string to be used when the object is printed"""
        return "Quasispecies Cell %d Type %d (%s)" % (self.id, self.type, self.types[self.type])

    def type(self):
        """Return the name of the type of this cell"""
        return self.types[self.type]

    def update(self, neighbors):
        """ Update the cell based on its neighbors

        Empty cells will be replaced by a neighbor proportional to their fitness
        using roulette wheel selection. 
        
        Parameters:

        *neighbors*
            A list of neighboring cells

        """

        if self.type == self.EMPTY:
            parent = self.choose_neighbor(neighbors)
            self.type = parent.type
            
            #if we're not staying empty, mutate
            if self.type != self.EMPTY:
                self.genotype = self.mutate(parent.genotype)
                #and update type to reflect the new genotype
                self.type = self.genotype[0]+1
            self.world.update_type_count(self.EMPTY, self.type)
        else:
            #check if we should die
            if random.random() < self.death_rate:
                self.world.update_type_count(self.type, self.EMPTY)
                self.type = self.EMPTY
                



# -*- coding: utf-8 -*-
""" NormalResource ResourceCell represents resources that have some initial
level, which increases and decreases through inflow and decay, respectively.
Additionally, resources can flow between neighboring nodes through diffusion.
"""

__author__ = "Brian Connelly <bdc@msu.edu>"
__credits__ = "Brian Connelly"

from seeds.Action import *
from seeds.ResourceCell import *
from seeds.SEEDSError import *
from seeds.utils.parsing import parse_int_rangelist


class NormalResource(ResourceCell):
    """Environmental Resource class

    Properties:

    id
        A unique ID for this node in the resource graph
    experiment
        A reference to the Experiment being performed
    resource
        A reference to the Resource of which this object is a part
    level
        The current level of the resource
    config_section
        The name of the section where parameters for this object are defined
    inflow
        Amount of resource (in units) that flows into the environment per epoch
        (default: 0.0)
    decay
        Fraction of the resource that is removed from the environment per epoch
        (default: 0.0)
    diffusion
        Fraction of resource difference (in percent) that flows into
        neighboring resource cells per epoch.  When a node is updated, the
        levels of its neighbors are checked.  The node then distributes some of
        its resource to the neighboring nodes with lower levels of resource,
        with the neighbor with the lowest level having first priority.  The
            difference between the levels of the two nodes is computed.  This
            is the maximum amount of resource that can flow to the neighboring
            cell.  Diffusion is the percentage of this difference that is
            transferred.  At 0, no resource is transferred, while at 1, the
            entire difference (or as much as the node with higher resource has)
            between the two is transferred.  This would result in the
            neighboring node having a greater level than the focal node.  This
            value can be seen to loosely represent viscosity. (default: 0.5)
    initial
        The amount of resource (in units) present in the environment at the
        beginning (default: 0.0)

    NormalResource Resources can be defined in config files using a
    [Resource:<uniquename>] section and using the NormalResource type.  For
    example:

        [Resource:glucose]
        type = NormalResource
        topology = CartesianTopology
        initial = 0
        inflow = 0.24
        diffusion = 0.1
        decay = 0.1

    The effects of diffusion will depend on the topology, specifically the number
    of neighboring resource cells.

    """

    def __init__(self, experiment, resource, config_section, id):
        """ Initialize a NormalResource object

        Parameters:

        *experiment*
            A pointer to the Experiment
        *resource*
            A pointer to the Resource of which this is a part
        *config_section*
            The name under which the configuration parameters are specified for
            this Resource
        *id*
            A unique ID for this node in the resource graph

        """
        super(NormalResource, self).__init__(experiment=experiment,
                                             resource=resource,
                                             config_section=config_section,
                                             id=id)

        self.inflow = self.experiment.config.getfloat(self.config_section, "inflow", default=0.0)
        self.diffusion = self.experiment.config.getfloat(self.config_section, "diffusion", default=0.5)
        self.decay = self.experiment.config.getfloat(self.config_section, "decay", default=0.0)
        self.initial = self.experiment.config.getfloat(self.config_section, "initial", default=0.0)

        self.level = self.initial * 1.0

        if self.inflow < 0:
            raise ConfigurationError("NormalResource: inflow for '%s' can not be negative" % (self.resource.name))
        elif self.diffusion < 0:
            raise ConfigurationError("NormalResource: diffusion for '%s' can not be negative" % (self.resource.name))
        elif self.diffusion > 1:
            raise ConfigurationError("NormalResource: diffusion for '%s' can not be greater than 1" % (self.resource.name))
        elif self.decay < 0:
            raise ConfigurationError("NormalResource: decay for '%s' can not be negative" % (self.resource.name))
        elif self.decay > 1:
            raise ConfigurationError("NormalResource: decay for '%s' can not be greater than 1" % (self.resource.name))

    def __str__(self):
        """Produce a string to be used when a NormalResource object is printed"""
        return "NormalResource [Name: %s][Level: %f][Inflow: %f][Diffusion: %f][Decay: %f]" % (self.name, self.level, self.inflow, self.diffusion, self.decay)

    def set_inflow(self, value):
        """Set the inflow amount of the resource

        Parameters:

        *value*
            The new inflow amount of the resoruce

        """
        self.inflow = value

    def set_diffusion(self, value):
        """Set the diffusion rate of the resource

        Parameters:

        *value*
            The new diffusion rate of the resoruce

        """
        self.diffusion = value

    def set_decay(self, value):
        """Set the decay rate of the resource

        Parameters:

        *value*
            The new decay rate of the resoruce

        """
        self.decay = value

    def set_initial(self, value):
        """Set the initial level of the resource

        Parameters:

        *value*
            The new initial level of the resoruce

        """
        self.initial = value

    def update(self):
        """Update the level of the resource.  Even if the resource is
        unavailable, the resource level will be updated.
        
        """
        neighbors = self.get_neighbors()

        # Adjust the level based on inflow and decay
        newlevel = (self.level * (1 - self.decay)) + self.inflow
        self.level = max(0, newlevel)

        # Find the neighbors with lower levels
        low_neighbors = []
        for n in neighbors:
            if n.level < self.level:
                low_neighbors.append(n)
        low_neighbors.sort()

        # Go through the neighboring nodes and transfer some resource to those
        # nodes as long as level is still above them.  Priority is given to
        # nodes with the lowest level.
        for n in low_neighbors:
            if self.level > n.level:
                xfer = min(self.level, self.level - n.level) * self.diffusion
                n.level += xfer
                self.level -= xfer

        self.level = max(0, newlevel)
        self.experiment.data['resources'][self.resource.name]['levels'][self.id] = self.level


class SetNormalResourceProperties(Action):
    """ Action to set the properties (inflow, decay, diffusion, or level) of a
    NormalResource resource

    Configuration is done in the [SetNormalResourceProperties] section

    Configuration Options:

    epoch_start
        The epoch at which to start executing (default: 0)
    epoch_end
        The epoch at which to stop executing (default: end of experiment)
    frequency
        The frequency (epochs) at which to execute (default: 1)
    priority
        The priority of this action.  Actions with higher priority get run
        first.  (default: 0)
    resource
        The name of the resource whose availability to set
    cells
        A list of the ResourceCell cells whose inflow to set.  If none are
        specified, the inflows all cells will be set.
    inflow
        The inflow value to be used.  If not specified, inflow will not be
        altered.
    decay
        The decay value to be used.  If not specified, decay will not be
        altered.
    diffusion
        The diffusion value to be used.  If not specified, diffusion will not
        be altered.
    level
        The level value to be used.  If not specified, level will not
        be altered.


    Configuration Example:

    [SetNormalResourceProperties]
    resource = glucose
    inflow = 10.2
    epoch_start = 0
    epoch_end = 100
    frequency = 1
    priority = 0

    """

    def __init__(self, experiment, label=None):
        """Initialize the SetNormalResourceProperties Action"""

        super(SetNormalResourceProperties, self).__init__(experiment,
                                                name="SetNormalResourceProperties",
                                                label=label)

        self.threshold = self.experiment.config.getint(self.config_section, 'threshold', 0)
        self.epoch_start = self.experiment.config.getint(self.config_section, 'epoch_start', 0)
        self.epoch_end = self.experiment.config.getint(self.config_section, 'epoch_end',
                                                  default=self.experiment.config.getint('Experiment', 'epochs', default=-1))
        self.frequency = self.experiment.config.getint(self.config_section, 'frequency', 1)
        self.priority = self.experiment.config.getint(self.config_section, 'priority', 0)

        self.resource = self.experiment.config.get(self.config_section, 'resource')

        try:
            self.res = self.experiment.resources[self.resource]
        except KeyError:
            raise ConfigurationError("SetNormalResourceProperties: Resource '%s' is undefined" % (self.resource))

        if self.res.type != 'NormalResource':
            raise ConfigurationError("SetNormalResourceProperties: Resource '%s' is not a NormalResource" % (self.resource))
            

        self.inflow = self.experiment.config.getfloat(self.config_section, 'inflow')
        if self.inflow and self.inflow < 0:
            raise ConfigurationError("SetNormalResourceProperties: Invalid value for inflow '%f'.  Must be nonnegative." % (self.inflow))


        self.decay = self.experiment.config.getfloat(self.config_section, 'decay')
        if self.decay and self.decay < 0:
            raise ConfigurationError("SetNormalResourceProperties: Invalid value for decay '%f'.  Must be nonnegative." % (self.decay))
        elif self.decay and self.decay > 1:
            raise ConfigurationError("SetNormalResourceProperties: Invalid value for decay '%f'.  Must not be greater than 1." % (self.decay))

        self.diffusion = self.experiment.config.getfloat(self.config_section, 'diffusion')
        if self.diffusion and self.diffusion < 0:
            raise ConfigurationError("SetNormalResourceProperties: Invalid value for diffusion '%f'.  Must be nonnegative." % (self.diffusion))
        elif self.diffusion and self.diffusion > 1:
            raise ConfigurationError("SetNormalResourceProperties: Invalid value for diffusion '%f'.  Must not be greater than 1." % (self.diffusion))

        self.level = self.experiment.config.getfloat(self.config_section, 'level')
        if self.level and self.level < 0:
            raise ConfigurationError("SetNormalResourceProperties: Invalid value for level '%f'.  Must be nonnegative." % (self.level))

        if not self.inflow and not self.decay and not self.diffusion and not self.level:
            raise ConfigurationError("SetNormalResourceProperties: Must specify value for inflow, decay, diffusion, or level")

        self.cells_str = self.experiment.config.get(self.config_section, 'cells')

        if not self.cells_str:
            self.cells = self.res.topology.graph.nodes()
        else:
            self.cells = parse_int_rangelist(self.cells_str, sorted=True)
            nids = self.res.topology.graph.nodes()
            for c in self.cells:
                if c not in nids:
                    raise ConfigurationError("SetNormalResourceProperties: Cell %d does not exist in Resource '%s'" % (c, self.resource))

    def update(self):
        """Execute the Action"""
        if self.skip_update():
	        return

        for c in self.cells:
            if self.inflow:
                self.res.topology.graph.node[c]['resource'].inflow = self.inflow
            if self.diffusion:
                self.res.topology.graph.node[c]['resource'].diffusion = self.diffusion
            if self.decay:
                self.res.topology.graph.node[c]['resource'].decay = self.decay
            if self.level:
                self.res.topology.graph.node[c]['resource'].level = self.level
